// EvCaller.cpp : Implementation of CEventCaller
#include "stdafx.h"
#include "EventCaller.h"
#include "EvCaller.h"
#include <algorithm>

using namespace std;

LPCWSTR pszConnStr = L"DSN=EventCaller;PWD=e4rtVaf;UID=sa;Database=EventCaller";
LPCWSTR pszCmd = L"select * from schedule where rtime <= (select rtime from schedule where signal='RELOAD_SCHEDULE') order by rtime asc";
LPCWSTR pszResetEvent = L"RELOAD_SCHEDULE";
const long c_100nsPerSec = 10000000;

/////////////////////////////////////////////////////////////////////////////
// CEventCaller
CEventCaller::CEventCaller()
{
	m_bStarted = false;
	m_bLeaveThreads = false;
	m_hEventThread = NULL;
}

HRESULT CEventCaller::FinalConstruct()
{
	HRESULT hr = S_OK;
	try
	{
		InitializeCriticalSection(&m_cs);
		EnterCriticalSection(&m_cs);

/*		DWORD dwThreadId = 0;
		m_hEventThread = CreateThread(NULL, 0, EventFunc
			, this, 0, &dwThreadId);
		if (!m_hEventThread)
			hr = MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WIN32, GetLastError());*/
	}
	catch(...)
	{
		hr = E_OUTOFMEMORY;
	}
	return hr;
}

void CEventCaller::FinalRelease()
{
	//  Let worker threads terminate itself.
	m_bLeaveThreads = true;
	if (!m_bStarted)
		LeaveCriticalSection(&m_cs);
	if (m_hEventThread != NULL)
		WaitForSingleObject(m_hEventThread, 3000);
	CloseHandle(m_hEventThread);
}

////////////////////////////////////////////////////////////////
// Reload event settings from DB.
HRESULT CEventCaller::Reset(strv_t &signals, longv_t &rtimes)
{
	HRESULT hr = S_OK;
	CComPtr<_Connection> pCnn;
	CComPtr<_Recordset> pRs;
	try
	{
		hr = pCnn.CoCreateInstance(__uuidof(Connection));
		if (FAILED(hr))
			_com_issue_error(hr);
		ATLASSERT(pCnn != NULL);
		hr = pCnn->Open(CComBSTR(m_sConnectionString), NULL, NULL, 0);
		if (SUCCEEDED(hr))
			hr = pCnn->Execute(CComBSTR(pszCmd), NULL, 0, &pRs);
		if (FAILED(hr))
			_com_issue_error(hr);

		// Copy event settings from recordset.
		while (1)
		{
			VARIANT_BOOL bEOF;
			hr = pRs->get_rsEOF(&bEOF);
			if (FAILED(hr) || bEOF)
				break;

			CComPtr<Fields> pFields;
			hr = pRs->get_Fields(&pFields);
			if (FAILED(hr))
				break;

			// signal field.
			CComVariant vItem(0), vSignal;
			CComPtr<Field> pField;
			hr = pFields->get_Item(vItem, &pField);
			if (SUCCEEDED(hr))
				hr = pField->get_Value(&vSignal);
			if (SUCCEEDED(hr))
				hr = vSignal.ChangeType(VT_BSTR);
			if (FAILED(hr))
				break;

			// rtime field.
			CComVariant vRTime;
			vItem = 1;
			pField.Release();
			hr = pFields->get_Item(vItem, &pField);
			if (SUCCEEDED(hr))
				hr = pField->get_Value(&vRTime);
			if (SUCCEEDED(hr))
				hr = vRTime.ChangeType(VT_I4);
			if (FAILED(hr))
				break;

			signals.push_back(wstring(vSignal.bstrVal));
			rtimes.push_back(vRTime.lVal);

			hr = pRs->MoveNext();
			if (FAILED(hr))
				break;
		}
	}
	catch (_com_error e)      
	{
		hr = e.Error();
	}
	catch (...)      
	{
		hr = E_FAIL;
	}

	if (pRs != NULL)
	{
		pRs->Close();
		pRs.Release();
	}
	if (pCnn != NULL)
	{
		pCnn->Close();
		pCnn.Release();
	}
	return hr;
}

class CTimeSlices
{
public:
	typedef CEventCaller::strv_t strv_t;
	typedef CEventCaller::longv_t longv_t;

	class CNode
	{
	public:
		CNode() { m_rtime = -1; m_sigref = -1; }
		CNode(long rtime, long sigref) { m_rtime = rtime; m_sigref = sigref; }

		bool operator<(const CNode &nd)
		{
			return m_rtime < nd.m_rtime;
		}

		long m_rtime;
		long m_sigref;
	};

	CTimeSlices(strv_t &signals, longv_t &rtimes)
	{
		// find RELOAD_SCHEDULE.
		for (int iReload = 0;iReload < signals.size();iReload++)
			if (signals[iReload] == pszResetEvent)
				break;
		ATLASSERT(iReload < signals.size());// this signal must be here.
		// fill time slices.
		for (int i = 0;i < signals.size();i++)
		{
			long rtime = rtimes[i];
			do
			{
				CNode nd(rtime, i);
				m_slices.push_back(nd);
				rtime += rtimes[i];
			} while (rtime <= rtimes[iReload]);
		}
		// order nodes by rtime.
		sort(m_slices.begin(), m_slices.end());
	}

	typedef vector< CNode > nodev_t;
	nodev_t m_slices;
};

// Execute one event cycle - up to RELOAD_SCHEDULE event.
HRESULT CEventCaller::RaiseEvents()
{
	strv_t signals;// signal names.
	longv_t rtimes;// corresponding timer handles.
	HRESULT hr = Reset(signals, rtimes);
	if (FAILED(hr))
		return hr;

	USES_CONVERSION;
	HANDLE hTimer = NULL;
	try
	{
		CTimeSlices ts(signals, rtimes);
		rtimes.clear();

		hTimer = CreateWaitableTimer(NULL, FALSE, NULL);
		if (hTimer == NULL)
			_com_issue_error(MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WIN32, GetLastError()));

		// loop by events.
		int iSlice = 0;
#ifdef _DEBUG
		DWORD dwStartTick = GetTickCount();
#endif // _DEBUG
		FILETIME ft;
		GetSystemTimeAsFileTime(&ft);
		__int64 i64 = (__int64(ft.dwHighDateTime) << 32) 
			+ __int64(ft.dwLowDateTime);
		while (1)
		{
			EnterCriticalSection(&m_cs);
			LeaveCriticalSection(&m_cs);
			if (m_bLeaveThreads)
				break;
			if (!m_bStarted)
			{
				m_bStarted = true;

				// Skip missed events
				GetSystemTimeAsFileTime(&ft);
				__int64 i64Delta = ((__int64(ft.dwHighDateTime) << 32) 
					+ __int64(ft.dwLowDateTime) - i64) / c_100nsPerSec;
				long ltm = i64Delta & 0xffffffff;
				bool bResetFound = false;
				for (;iSlice < ts.m_slices.size() && ltm > ts.m_slices[iSlice].m_rtime;iSlice++)
				{
					if (signals[ts.m_slices[iSlice].m_sigref] == wstring(pszResetEvent))
						bResetFound = true;
				}
				if (bResetFound)
					break;
			}

			long ltm = ts.m_slices[iSlice].m_rtime;
			__int64 i = i64 + ltm * c_100nsPerSec;
			LARGE_INTEGER liTime;
			liTime.HighPart = (LONG)(i >> 32);
			liTime.LowPart = (DWORD)(i & 0xffffffff);
			BOOL b = SetWaitableTimer(hTimer, &liTime, 0, NULL, NULL, FALSE);

			// Wait for timer.
			DWORD dw = WaitForSingleObject(hTimer, INFINITE);
#ifdef _DEBUG
			ATLTRACE(_T("Time=%li\n"), GetTickCount() - dwStartTick);
#endif // _DEBUG

			// raise events
			bool bResetFound = false;
			while (iSlice < ts.m_slices.size())
			{
				Fire_OnEvent(CComBSTR(signals[ts.m_slices[iSlice].m_sigref].data()));
				if (signals[ts.m_slices[iSlice].m_sigref] == wstring(pszResetEvent))
					bResetFound = true;
				iSlice++;
				if (ltm < ts.m_slices[iSlice].m_rtime)
					break;
			}
			if (bResetFound)
				break;
			ATLASSERT(iSlice < ts.m_slices.size());
		}
	}
	catch(_com_error e)
	{
		hr = e.Error();
	}
	catch(...)
	{
		hr = E_FAIL;
	}

	if (hTimer != NULL)
		CloseHandle(hTimer);

	return hr;
}

// Event thread function.
DWORD WINAPI CEventCaller::EventFunc(LPVOID pParam)
{
//	BOOL bRet = SetThreadPriority(GetCurrentThread()
//		, THREAD_PRIORITY_BELOW_NORMAL);
//	ATLASSERT(bRet);
	CEventCaller *pThis = (CEventCaller*)pParam;

	HRESULT hr = CoInitializeEx(NULL, COINIT_APARTMENTTHREADED);
	if (FAILED(hr))
		return hr;
	while (SUCCEEDED(hr))
	{
		// Wait for raising event started.
		if (pThis->m_bLeaveThreads)
			break;

		// Raise events.
		hr = pThis->RaiseEvents();
	}

	CoUninitialize();
	return hr;
}

// IEventCaller --------------------------------------------
HRESULT CEventCaller::Start()
{
	if (m_bStarted)
		return S_OK;

	HRESULT hr = S_OK;
	if (!m_hEventThread)
	{
		try
		{
			DWORD dwThreadId = 0;
			m_hEventThread = CreateThread(NULL, 0, EventFunc
				, this, 0, &dwThreadId);
			if (!m_hEventThread)
				hr = MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WIN32, GetLastError());
		}
		catch(...)
		{
			hr = E_OUTOFMEMORY;
		}
	}
	
	LeaveCriticalSection(&m_cs);
	return hr;
}

HRESULT CEventCaller::Stop()
{
	if (!m_bStarted)
		return S_OK;
	m_bStarted = false;
	EnterCriticalSection(&m_cs);
	return S_OK;
}

STDMETHODIMP CEventCaller::SetConnectionString(BSTR ConnectionString)
{
	m_sConnectionString = ConnectionString;
	return 0;
}
